@@ -7,18 +7,21 @@
 #include "foreign.h"
 #include "foreign_p.h"
 
-#include <QtQml/qqml.h>
-#include <QtQml/qqmlcomponent.h>
-#include <QtCore/qproperty.h>
-#include <QtCore/qtimeline.h>
-#include <QtCore/qrect.h>
 #include <QtQmlTypeRegistrar/private/qqmltyperegistrar_p.h>
-#include <QtCore/qtemporaryfile.h>
 
 #ifdef QT_QUICK_LIB
 #    include <QtQuick/qquickitem.h>
 #endif
 
+#include <QtQml/qqml.h>
+#include <QtQml/qqmlcomponent.h>
+
+#include <QtCore/qabstractitemmodel.h>
+#include <QtCore/qproperty.h>
+#include <QtCore/qrect.h>
+#include <QtCore/qtemporaryfile.h>
+#include <QtCore/qtimeline.h>
+
 class Interface {};
 class Interface2 {};
 class Interface3 {};
@@ -182,7 +185,7 @@ class DerivedFromForeign : public QTimeLine
     Q_OBJECT
     QML_ELEMENT
 public:
-    DerivedFromForeign(QObject *parent) : QTimeLine(1000, parent) {}
+    DerivedFromForeign(QObject *parent = nullptr) : QTimeLine(1000, parent) { }
 };
 
 class ExtensionA : public QObject
@@ -501,6 +504,18 @@ signals:
     void clonedSignal(int i = 7);
 };
 
+class Constructible
+{
+    Q_GADGET
+    QML_VALUE_TYPE(constructible)
+    QML_CONSTRUCTIBLE_VALUE
+public:
+    Q_INVOKABLE Constructible(int i = 12) : m_i(i) {}
+
+private:
+    int m_i;
+};
+
 class AnonymousAndUncreatable : public QObject
 {
      Q_OBJECT
@@ -519,6 +534,110 @@ struct InvisibleForeign
     QML_NAMED_ELEMENT(Invisible)
 };
 
+class TypedEnum : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+public:
+    enum S: qint16 {
+        A, B, C
+    };
+    Q_ENUM(S)
+
+    enum T: quint16 {
+        D, E, F
+    };
+    Q_ENUM(T)
+
+    enum U: qint8 {
+        G, H, I
+    };
+    Q_ENUM(U)
+
+    enum V: quint8 {
+        J, K, L
+    };
+    Q_ENUM(V)
+};
+
+class ListSignal : public QObject
+{
+    Q_OBJECT
+    QML_ANONYMOUS
+
+Q_SIGNALS:
+    void objectListHappened(const QList<QObject *> &);
+};
+
+class Bar : public QObject
+{
+    Q_OBJECT
+    Q_PROPERTY(int outerBarProp READ bar CONSTANT)
+public:
+    Bar(QObject *parent = nullptr) : QObject(parent) {}
+    int bar() const { return 44; }
+};
+
+namespace Testing {
+
+class Foo : public QObject
+{
+    Q_OBJECT
+    Q_PROPERTY(int fooProp READ foo CONSTANT)
+
+public:
+    int foo() const { return 42; }
+};
+
+class Bar : public Foo
+{
+    Q_OBJECT
+    QML_ELEMENT
+    Q_PROPERTY(int barProp READ bar CONSTANT)
+
+public:
+    int bar() const { return 43; }
+};
+
+namespace Inner {
+
+class Baz : public Bar
+{
+    Q_OBJECT
+    QML_ELEMENT
+
+    QML_EXTENDED(::Bar)
+    QML_ATTACHED(Foo)
+
+public:
+    static Foo *qmlAttachedProperties(QObject *) { return new Foo; }
+};
+
+} // namespace Inner
+} // namespace Testing
+
+struct QByteArrayStdVectorForeign
+{
+    Q_GADGET
+    QML_ANONYMOUS
+    QML_SEQUENTIAL_CONTAINER(QByteArray)
+    QML_FOREIGN(std::vector<QByteArray>)
+};
+
+// Anonymous value type for an unknown foreign type
+struct QPersistentModelIndexValueType
+{
+    QPersistentModelIndex v;
+    Q_PROPERTY(int row READ row FINAL)
+    Q_GADGET
+    QML_ANONYMOUS
+    QML_EXTENDED(QPersistentModelIndexValueType)
+    QML_FOREIGN(QPersistentModelIndex)
+
+public:
+    inline int row() const { return v.row(); }
+};
+
 class tst_qmltyperegistrar : public QObject
 {
     Q_OBJECT
@@ -567,8 +686,14 @@ private slots:
     void duplicateExportWarnings();
     void clonedSignal();
     void baseVersionInQmltypes();
+    void constructibleValueType();
     void anonymousAndUncreatable();
     void omitInvisible();
+    void typedEnum();
+    void listSignal();
+    void withNamespace();
+    void sequenceRegistration();
+    void valueTypeSelfReference();
 
 private:
     QByteArray qmltypesData;
