@@ -56,6 +56,10 @@ private slots:
     void sectionGeometryChange();
     void areaZeroviewDoesNotNeedlesslyPopulateWholeModel();
 
+    void delegateContextHandling();
+    void fetchMore_data();
+    void fetchMore();
+
 private:
     void flickWithTouch(QQuickWindow *window, const QPoint &from, const QPoint &to);
     QScopedPointer<QPointingDevice> touchDevice = QScopedPointer<QPointingDevice>(QTest::createTouchDevice());
@@ -384,29 +388,31 @@ void tst_QQuickListView2::tapDelegateDuringFlicking_data()
 {
     QTest::addColumn<QByteArray>("qmlFile");
     QTest::addColumn<QQuickFlickable::BoundsBehavior>("boundsBehavior");
+    QTest::addColumn<bool>("expectCanceled");
 
     QTest::newRow("Button StopAtBounds") << QByteArray("buttonDelegate.qml")
-                                         << QQuickFlickable::BoundsBehavior(QQuickFlickable::StopAtBounds);
+                                         << QQuickFlickable::BoundsBehavior(QQuickFlickable::StopAtBounds) << false;
     QTest::newRow("MouseArea StopAtBounds") << QByteArray("mouseAreaDelegate.qml")
-                                            << QQuickFlickable::BoundsBehavior(QQuickFlickable::StopAtBounds);
+                                            << QQuickFlickable::BoundsBehavior(QQuickFlickable::StopAtBounds) << true;
     QTest::newRow("Button DragOverBounds") << QByteArray("buttonDelegate.qml")
-                                           << QQuickFlickable::BoundsBehavior(QQuickFlickable::DragOverBounds);
+                                           << QQuickFlickable::BoundsBehavior(QQuickFlickable::DragOverBounds) << false;
     QTest::newRow("MouseArea DragOverBounds") << QByteArray("mouseAreaDelegate.qml")
-                                              << QQuickFlickable::BoundsBehavior(QQuickFlickable::DragOverBounds);
+                                              << QQuickFlickable::BoundsBehavior(QQuickFlickable::DragOverBounds) << true;
     QTest::newRow("Button OvershootBounds") << QByteArray("buttonDelegate.qml")
-                                            << QQuickFlickable::BoundsBehavior(QQuickFlickable::OvershootBounds);
+                                            << QQuickFlickable::BoundsBehavior(QQuickFlickable::OvershootBounds) << false;
     QTest::newRow("MouseArea OvershootBounds") << QByteArray("mouseAreaDelegate.qml")
-                                               << QQuickFlickable::BoundsBehavior(QQuickFlickable::OvershootBounds);
+                                               << QQuickFlickable::BoundsBehavior(QQuickFlickable::OvershootBounds) << true;
     QTest::newRow("Button DragAndOvershootBounds") << QByteArray("buttonDelegate.qml")
-                                                   << QQuickFlickable::BoundsBehavior(QQuickFlickable::DragAndOvershootBounds);
+                                                   << QQuickFlickable::BoundsBehavior(QQuickFlickable::DragAndOvershootBounds) << false;
     QTest::newRow("MouseArea DragAndOvershootBounds") << QByteArray("mouseAreaDelegate.qml")
-                                                      << QQuickFlickable::BoundsBehavior(QQuickFlickable::DragAndOvershootBounds);
+                                                      << QQuickFlickable::BoundsBehavior(QQuickFlickable::DragAndOvershootBounds) << true;
 }
 
 void tst_QQuickListView2::tapDelegateDuringFlicking() // QTBUG-103832
 {
     QFETCH(QByteArray, qmlFile);
     QFETCH(QQuickFlickable::BoundsBehavior, boundsBehavior);
+    QFETCH(bool, expectCanceled);
 
     QQuickView window;
     QVERIFY(QQuickTest::showView(window, testFileUrl(qmlFile.constData())));
@@ -419,7 +425,16 @@ void tst_QQuickListView2::tapDelegateDuringFlicking() // QTBUG-103832
     QVERIFY(listView->isFlicking()); // we want to test the case when it's still moving while we tap
     // @y = 400 we pressed the 4th delegate; started flicking, and the press was canceled
     QCOMPARE(listView->property("pressedDelegates").toList().first(), 4);
-    QCOMPARE(listView->property("canceledDelegates").toList().first(), 4);
+    // At first glance one would expect MouseArea and Button would be consistent about this;
+    // but in fact, before ListView takes over the grab via filtering,
+    // Button.pressed transitions to false because QQuickAbstractButtonPrivate::handleMove
+    // sees that the touchpoint has strayed outside its bounds, but it does NOT emit the canceled signal
+    if (expectCanceled) {
+        const QVariantList canceledDelegates = listView->property("canceledDelegates").toList();
+        QCOMPARE(canceledDelegates.size(), 1);
+        QCOMPARE(canceledDelegates.first(), 4);
+    }
+    QCOMPARE(listView->property("releasedDelegates").toList().size(), 0);
 
     // press a delegate during flicking (at y > 501 + 100, so likely delegate 6)
     QTest::touchEvent(&window, touchDevice.data()).press(0, {100, 100});
@@ -442,7 +457,7 @@ void tst_QQuickListView2::tapDelegateDuringFlicking() // QTBUG-103832
     QVERIFY(lastPressed > 5);
     QCOMPARE(releasedDelegates.last(), lastPressed);
     QCOMPARE(tappedDelegates.last(), lastPressed);
-    QCOMPARE(canceledDelegates.size(), 1); // only the first press was canceled, not the second
+    QCOMPARE(canceledDelegates.size(), expectCanceled ? 1 : 0); // only the first press was canceled, not the second
 }
 
 void tst_QQuickListView2::flickDuringFlicking_data()
@@ -701,122 +716,122 @@ void tst_QQuickListView2::wheelSnap_data()
 
     QTest::newRow("vertical, top to bottom")
             << QQuickListView::Vertical << Qt::LeftToRight << QQuickItemView::TopToBottom
-            << QQuickItemView::NoHighlightRange << QPoint(20, -120) << 200.0 << 600.0 << 0.0 << 0.0
+            << QQuickItemView::NoHighlightRange << QPoint(20, -240) << 200.0 << 600.0 << 0.0 << 0.0
             << 0.0;
 
     QTest::newRow("vertical, bottom to top")
             << QQuickListView::Vertical << Qt::LeftToRight << QQuickItemView::BottomToTop
-            << QQuickItemView::NoHighlightRange << QPoint(20, 120) << -400.0 << -800.0 << -200.0
+            << QQuickItemView::NoHighlightRange << QPoint(20, 240) << -400.0 << -800.0 << -200.0
             << 0.0 << 0.0;
 
     QTest::newRow("horizontal, left to right")
             << QQuickListView::Horizontal << Qt::LeftToRight << QQuickItemView::TopToBottom
-            << QQuickItemView::NoHighlightRange << QPoint(-120, 20) << 200.0 << 600.0 << 0.0 << 0.0
+            << QQuickItemView::NoHighlightRange << QPoint(-240, 20) << 200.0 << 600.0 << 0.0 << 0.0
             << 0.0;
 
     QTest::newRow("horizontal, right to left")
             << QQuickListView::Horizontal << Qt::RightToLeft << QQuickItemView::TopToBottom
-            << QQuickItemView::NoHighlightRange << QPoint(120, 20) << -400.0 << -800.0 << -200.0
+            << QQuickItemView::NoHighlightRange << QPoint(240, 20) << -400.0 << -800.0 << -200.0
             << 0.0 << 0.0;
 
     QTest::newRow("vertical, top to bottom, enforce range")
             << QQuickListView::Vertical << Qt::LeftToRight << QQuickItemView::TopToBottom
-            << QQuickItemView::StrictlyEnforceRange << QPoint(20, -120) << 200.0 << 600.0 << 0.0
+            << QQuickItemView::StrictlyEnforceRange << QPoint(20, -240) << 200.0 << 600.0 << 0.0
             << 0.0 << 0.0;
 
     QTest::newRow("vertical, bottom to top, enforce range")
             << QQuickListView::Vertical << Qt::LeftToRight << QQuickItemView::BottomToTop
-            << QQuickItemView::StrictlyEnforceRange << QPoint(20, 120) << -400.0 << -800.0 << -200.0
+            << QQuickItemView::StrictlyEnforceRange << QPoint(20, 240) << -400.0 << -800.0 << -200.0
             << 0.0 << 0.0;
 
     QTest::newRow("horizontal, left to right, enforce range")
             << QQuickListView::Horizontal << Qt::LeftToRight << QQuickItemView::TopToBottom
-            << QQuickItemView::StrictlyEnforceRange << QPoint(-120, 20) << 200.0 << 600.0 << 0.0
+            << QQuickItemView::StrictlyEnforceRange << QPoint(-240, 20) << 200.0 << 600.0 << 0.0
             << 0.0 << 0.0;
 
     QTest::newRow("horizontal, right to left, enforce range")
             << QQuickListView::Horizontal << Qt::RightToLeft << QQuickItemView::TopToBottom
-            << QQuickItemView::StrictlyEnforceRange << QPoint(120, 20) << -400.0 << -800.0 << -200.0
+            << QQuickItemView::StrictlyEnforceRange << QPoint(240, 20) << -400.0 << -800.0 << -200.0
             << 0.0 << 0.0;
 
     QTest::newRow("vertical, top to bottom, apply range")
             << QQuickListView::Vertical << Qt::LeftToRight << QQuickItemView::TopToBottom
-            << QQuickItemView::ApplyRange << QPoint(20, -120) << 200.0 << 600.0 << 0.0 << 0.0
+            << QQuickItemView::ApplyRange << QPoint(20, -240) << 200.0 << 600.0 << 0.0 << 0.0
             << 0.0;
 
     QTest::newRow("vertical, bottom to top, apply range")
             << QQuickListView::Vertical << Qt::LeftToRight << QQuickItemView::BottomToTop
-            << QQuickItemView::ApplyRange << QPoint(20, 120) << -400.0 << -800.0 << -200.0 << 0.0
+            << QQuickItemView::ApplyRange << QPoint(20, 240) << -400.0 << -800.0 << -200.0 << 0.0
             << 0.0;
 
     QTest::newRow("horizontal, left to right, apply range")
             << QQuickListView::Horizontal << Qt::LeftToRight << QQuickItemView::TopToBottom
-            << QQuickItemView::ApplyRange << QPoint(-120, 20) << 200.0 << 600.0 << 0.0 << 0.0
+            << QQuickItemView::ApplyRange << QPoint(-240, 20) << 200.0 << 600.0 << 0.0 << 0.0
             << 0.0;
 
     QTest::newRow("horizontal, right to left, apply range")
             << QQuickListView::Horizontal << Qt::RightToLeft << QQuickItemView::TopToBottom
-            << QQuickItemView::ApplyRange << QPoint(120, 20) << -400.0 << -800.0 << -200.0 << 0.0
+            << QQuickItemView::ApplyRange << QPoint(240, 20) << -400.0 << -800.0 << -200.0 << 0.0
             << 0.0;
 
     QTest::newRow("vertical, top to bottom with highlightRange")
             << QQuickListView::Vertical << Qt::LeftToRight << QQuickItemView::TopToBottom
-            << QQuickItemView::NoHighlightRange << QPoint(20, -120) << 190.0 << 600.0 << 0.0 << 10.0
+            << QQuickItemView::NoHighlightRange << QPoint(20, -240) << 190.0 << 600.0 << 0.0 << 10.0
             << 210.0;
 
     QTest::newRow("vertical, bottom to top with highlightRange")
             << QQuickListView::Vertical << Qt::LeftToRight << QQuickItemView::BottomToTop
-            << QQuickItemView::NoHighlightRange << QPoint(20, 120) << -390.0 << -800.0 << -200.0
+            << QQuickItemView::NoHighlightRange << QPoint(20, 240) << -390.0 << -800.0 << -200.0
             << 10.0 << 210.0;
 
     QTest::newRow("horizontal, left to right with highlightRange")
             << QQuickListView::Horizontal << Qt::LeftToRight << QQuickItemView::TopToBottom
-            << QQuickItemView::NoHighlightRange << QPoint(-120, 20) << 190.0 << 600.0 << 0.0 << 10.0
+            << QQuickItemView::NoHighlightRange << QPoint(-240, 20) << 190.0 << 600.0 << 0.0 << 10.0
             << 210.0;
 
     QTest::newRow("horizontal, right to left with highlightRange")
             << QQuickListView::Horizontal << Qt::RightToLeft << QQuickItemView::TopToBottom
-            << QQuickItemView::NoHighlightRange << QPoint(120, 20) << -390.0 << -800.0 << -200.0
+            << QQuickItemView::NoHighlightRange << QPoint(240, 20) << -390.0 << -800.0 << -200.0
             << 10.0 << 210.0;
 
     QTest::newRow("vertical, top to bottom, enforce range with highlightRange")
             << QQuickListView::Vertical << Qt::LeftToRight << QQuickItemView::TopToBottom
-            << QQuickItemView::StrictlyEnforceRange << QPoint(20, -120) << 190.0 << 590.0 << -10.0
+            << QQuickItemView::StrictlyEnforceRange << QPoint(20, -240) << 190.0 << 590.0 << -10.0
             << 10.0 << 210.0;
 
     QTest::newRow("vertical, bottom to top, enforce range with highlightRange")
             << QQuickListView::Vertical << Qt::LeftToRight << QQuickItemView::BottomToTop
-            << QQuickItemView::StrictlyEnforceRange << QPoint(20, 120) << -390.0 << -790.0 << -190.0
+            << QQuickItemView::StrictlyEnforceRange << QPoint(20, 240) << -390.0 << -790.0 << -190.0
             << 10.0 << 210.0;
 
     QTest::newRow("horizontal, left to right, enforce range with highlightRange")
             << QQuickListView::Horizontal << Qt::LeftToRight << QQuickItemView::TopToBottom
-            << QQuickItemView::StrictlyEnforceRange << QPoint(-120, 20) << 190.0 << 590.0 << -10.0
+            << QQuickItemView::StrictlyEnforceRange << QPoint(-240, 20) << 190.0 << 590.0 << -10.0
             << 10.0 << 210.0;
 
     QTest::newRow("horizontal, right to left, enforce range with highlightRange")
             << QQuickListView::Horizontal << Qt::RightToLeft << QQuickItemView::TopToBottom
-            << QQuickItemView::StrictlyEnforceRange << QPoint(120, 20) << -390.0 << -790.0 << -190.0
+            << QQuickItemView::StrictlyEnforceRange << QPoint(240, 20) << -390.0 << -790.0 << -190.0
             << 10.0 << 210.0;
 
     QTest::newRow("vertical, top to bottom, apply range with highlightRange")
             << QQuickListView::Vertical << Qt::LeftToRight << QQuickItemView::TopToBottom
-            << QQuickItemView::ApplyRange << QPoint(20, -120) << 190.0 << 600.0 << 0.0 << 10.0
+            << QQuickItemView::ApplyRange << QPoint(20, -240) << 190.0 << 600.0 << 0.0 << 10.0
             << 210.0;
 
     QTest::newRow("vertical, bottom to top, apply range with highlightRange")
             << QQuickListView::Vertical << Qt::LeftToRight << QQuickItemView::BottomToTop
-            << QQuickItemView::ApplyRange << QPoint(20, 120) << -390.0 << -800.0 << -200.0 << 10.0
+            << QQuickItemView::ApplyRange << QPoint(20, 240) << -390.0 << -800.0 << -200.0 << 10.0
             << 210.0;
 
     QTest::newRow("horizontal, left to right, apply range with highlightRange")
             << QQuickListView::Horizontal << Qt::LeftToRight << QQuickItemView::TopToBottom
-            << QQuickItemView::ApplyRange << QPoint(-120, 20) << 190.0 << 600.0 << 0.0 << 10.0
+            << QQuickItemView::ApplyRange << QPoint(-240, 20) << 190.0 << 600.0 << 0.0 << 10.0
             << 210.0;
 
     QTest::newRow("horizontal, right to left, apply range with highlightRange")
             << QQuickListView::Horizontal << Qt::RightToLeft << QQuickItemView::TopToBottom
-            << QQuickItemView::ApplyRange << QPoint(120, 20) << -390.0 << -800.0 << -200.0 << 10.0
+            << QQuickItemView::ApplyRange << QPoint(240, 20) << -390.0 << -800.0 << -200.0 << 10.0
             << 210.0;
 }
 
@@ -1011,6 +1026,96 @@ void tst_QQuickListView2::areaZeroviewDoesNotNeedlesslyPopulateWholeModel()
     QVERIFY(delegateCreationCounter() < 100);
 }
 
+void tst_QQuickListView2::delegateContextHandling()
+{
+    QQmlEngine engine;
+    QQmlComponent c(&engine, testFileUrl("delegateContextHandling.qml"));
+    QVERIFY2(c.isReady(), qPrintable(c.errorString()));
+    std::unique_ptr<QObject> o(c.create());
+    QVERIFY(o);
+
+    for (int i = 0; i < 10; ++i) {
+        QQuickItem *delegate = nullptr;
+        QMetaObject::invokeMethod(o.get(), "toggle", Q_RETURN_ARG(QQuickItem *, delegate));
+        QVERIFY(delegate);
+    }
+
+}
+
+class TestFetchMoreModel : public QAbstractListModel
+{
+    Q_OBJECT
+
+public:
+    QVariant data(const QModelIndex& index, int role) const override
+    {
+        if (role == Qt::DisplayRole)
+            return QString::number(index.row());
+        return {};
+    }
+
+    int columnCount(const QModelIndex&) const override { return 1; }
+
+    int rowCount(const QModelIndex& parent) const override
+    {
+        return parent.isValid() ? 0 : m_lines;
+    }
+
+    QModelIndex parent(const QModelIndex&) const override { return {}; }
+
+    bool canFetchMore(const QModelIndex &) const override { return true; }
+
+    void fetchMore(const QModelIndex & parent) override
+    {
+        if (Q_UNLIKELY(parent.isValid()))
+            return;
+        beginInsertRows(parent, m_lines, m_lines);
+        m_lines++;
+        endInsertRows();
+    }
+
+    int m_lines = 3;
+};
+
+void tst_QQuickListView2::fetchMore_data()
+{
+    QTest::addColumn<bool>("reuseItems");
+    QTest::addColumn<int>("cacheBuffer");
+
+    QTest::newRow("no reuseItems, default buffer") << false << -1;
+    QTest::newRow("reuseItems, default buffer") << true << -1;
+    QTest::newRow("no reuseItems, no buffer") << false << 0;
+    QTest::newRow("reuseItems, no buffer") << true << 0;
+    QTest::newRow("no reuseItems, buffer 100 px") << false << 100;
+    QTest::newRow("reuseItems, buffer 100 px") << true << 100;
+}
+
+void tst_QQuickListView2::fetchMore() // QTBUG-95107
+{
+    QFETCH(bool, reuseItems);
+    QFETCH(int, cacheBuffer);
+
+    TestFetchMoreModel model;
+    qmlRegisterSingletonInstance("org.qtproject.Test", 1, 0, "FetchMoreModel", &model);
+    QQuickView window;
+    QVERIFY(QQuickTest::showView(window, testFileUrl("fetchMore.qml")));
+    auto *listView = qobject_cast<QQuickListView*>(window.rootObject());
+    QVERIFY(listView);
+    listView->setReuseItems(reuseItems);
+    if (cacheBuffer >= 0)
+        listView->setCacheBuffer(cacheBuffer);
+
+    for (int i = 0; i < 3; ++i) {
+        const int rowCount = listView->count();
+        if (lcTests().isDebugEnabled()) QTest::qWait(1000);
+        listView->flick(0, -5000);
+        QTRY_VERIFY(!listView->isMoving());
+        qCDebug(lcTests) << "after flick: contentY" << listView->contentY()
+                         << "rows" << rowCount << "->" << listView->count();
+        QCOMPARE_GT(listView->count(), rowCount);
+        QCOMPARE_GE(model.m_lines, listView->count()); // fetchMore() was called
+    }
+}
 QTEST_MAIN(tst_QQuickListView2)
 
 #include "tst_qquicklistview2.moc"
